@@ -12,7 +12,8 @@ from .S3 import S3
 from .Config import Config
 from .S3Uri import S3Uri
 from .FileDict import FileDict
-from .Utils import *
+from .BaseUtils import dateS3toUnix, dateRFC822toUnix
+from .Utils import unicodise, deunicodise, deunicodise_s, replace_nonprintables
 from .Exceptions import ParameterError
 from .HashCache import HashCache
 
@@ -25,6 +26,8 @@ import re
 import errno
 import io
 
+PY3 = (sys.version_info >= (3, 0))
+
 __all__ = ["fetch_local_list", "fetch_remote_list", "compare_filelists"]
 
 def _os_walk_unicode(top):
@@ -33,7 +36,7 @@ def _os_walk_unicode(top):
     '''
     try:
         names = os.listdir(deunicodise(top))
-    except:
+    except Exception:
         return
 
     dirs, nondirs = [], []
@@ -57,34 +60,39 @@ def handle_exclude_include_walk_dir(root, dirname):
     Should this root/dirname directory be excluded? (otherwise included by default)
     Exclude dir matches in the current directory
     This prevents us from recursing down trees we know we want to ignore
-    return True for including, and False for excluding
+    return True for excluding, and False for including
     '''
     cfg = Config()
+    # python versions end their patterns (from globs) differently, test for different styles; check python3.6+ styles first
+    directory_patterns = (u'/)$', u'/)\\Z', u'\\/$', u'\\/\\Z(?ms)')
+
     d = os.path.join(root, dirname, '')
-    debug(u"CHECK: %r" % d)
+    debug(u"CHECK: '%s'" % d)
     excluded = False
     for r in cfg.exclude:
-        # python versions end their patterns (from globs) differently, test for both styles.
-        if not (r.pattern.endswith(u'\\/$') or r.pattern.endswith(u'\\/\\Z(?ms)')): continue # we only check for directory patterns here
+        if not any(r.pattern.endswith(dp) for dp in directory_patterns):
+            # we only check for directory patterns here
+            continue
         if r.search(d):
             excluded = True
-            debug(u"EXCL-MATCH: '%s'" % (cfg.debug_exclude[r]))
+            debug(u"EXCL-MATCH: '%s'" % cfg.debug_exclude[r])
             break
     if excluded:
         ## No need to check for --include if not excluded
         for r in cfg.include:
-            # python versions end their patterns (from globs) differently, test for both styles.
-            if not (r.pattern.endswith(u'\\/$') or r.pattern.endswith(u'\\/\\Z(?ms)')): continue # we only check for directory patterns here
-            debug(u"INCL-TEST: %s ~ %s" % (d, r.pattern))
+            if not any(r.pattern.endswith(dp) for dp in directory_patterns):
+                # we only check for directory patterns here
+                continue
+            debug(u"INCL-TEST: '%s' ~ %s" % (d, r.pattern))
             if r.search(d):
                 excluded = False
                 debug(u"INCL-MATCH: '%s'" % (cfg.debug_include[r]))
                 break
     if excluded:
         ## Still excluded - ok, action it
-        debug(u"EXCLUDE: %r" % d)
+        debug(u"EXCLUDE: '%s'" % d)
     else:
-        debug(u"PASS: %r" % d)
+        debug(u"PASS: '%s'" % d)
     return excluded
 
 def _fswalk_follow_symlinks(path):
@@ -125,28 +133,28 @@ def filter_exclude_include(src_list):
     cfg = Config()
     exclude_list = FileDict(ignore_case = False)
     for file in src_list.keys():
-        debug(u"CHECK: %s" % file)
+        debug(u"CHECK: '%s'" % file)
         excluded = False
         for r in cfg.exclude:
             if r.search(file):
                 excluded = True
-                debug(u"EXCL-MATCH: '%s'" % (cfg.debug_exclude[r]))
+                debug(u"EXCL-MATCH: '%s'" % cfg.debug_exclude[r])
                 break
         if excluded:
             ## No need to check for --include if not excluded
             for r in cfg.include:
                 if r.search(file):
                     excluded = False
-                    debug(u"INCL-MATCH: '%s'" % (cfg.debug_include[r]))
+                    debug(u"INCL-MATCH: '%s'" % cfg.debug_include[r])
                     break
         if excluded:
             ## Still excluded - ok, action it
-            debug(u"EXCLUDE: %s" % file)
+            debug(u"EXCLUDE: '%s'" % file)
             exclude_list[file] = src_list[file]
             del(src_list[file])
             continue
         else:
-            debug(u"PASS: %r" % (file))
+            debug(u"PASS: '%s'" % file)
     return src_list, exclude_list
 
 
@@ -182,9 +190,7 @@ def _get_filelist_from_file(cfg, local_path):
 
     # reformat to match os.walk()
     result = []
-    keys = filelist.keys()
-    keys.sort()
-    for key in keys:
+    for key in sorted(filelist):
         values = filelist[key]
         values.sort()
         result.append((key, [], values))
@@ -202,7 +208,8 @@ def fetch_local_list(args, is_src = False, recursive = None):
             if counter % 1000 == 0:
                 info(u"[%d/%d]" % (counter, len_loc_list))
 
-            if relative_file == '-': continue
+            if relative_file == '-':
+                continue
 
             full_name = loc_list[relative_file]['full_name']
             try:
@@ -243,7 +250,7 @@ def fetch_local_list(args, is_src = False, recursive = None):
             try:
                 uid = os.geteuid()
                 gid = os.getegid()
-            except:
+            except Exception:
                 uid = 0
                 gid = 0
             loc_list["-"] = {
@@ -306,18 +313,26 @@ def fetch_local_list(args, is_src = False, recursive = None):
         # not.  Leave it to a non-files_from run to purge.
         if cfg.cache_file and len(cfg.files_from) == 0:
             cache.mark_all_for_purge()
-            for i in local_list.keys():
-                cache.unmark_for_purge(local_list[i]['dev'], local_list[i]['inode'], local_list[i]['mtime'], local_list[i]['size'])
+            if PY3:
+                local_list_val_iter = local_list.values()
+            else:
+                local_list_val_iter = local_list.itervalues()
+            for f_info in local_list_val_iter:
+                inode = f_info.get('inode', 0)
+                if not inode:
+                    continue
+                cache.unmark_for_purge(f_info['dev'], inode, f_info['mtime'],
+                                       f_info['size'])
             cache.purge()
             cache.save(cfg.cache_file)
 
     cfg = Config()
 
     cache = HashCache()
-    if cfg.cache_file and os.path.exists(deunicodise_s(cfg.cache_file)):
+    if cfg.cache_file and os.path.isfile(deunicodise_s(cfg.cache_file)) and os.path.getsize(deunicodise_s(cfg.cache_file)) > 0:
         cache.load(cfg.cache_file)
     else:
-        info(u"No cache file found, creating it.")
+        info(u"Cache file not found or empty, creating/populating it.")
 
     local_uris = []
     local_list = FileDict(ignore_case = False)
@@ -362,7 +377,7 @@ def fetch_remote_list(args, require_attribs = False, recursive = None, uri_param
         remote_item.update({
         'size': int(response['headers']['content-length']),
         'md5': response['headers']['etag'].strip('"\''),
-        'timestamp' : dateRFC822toUnix(response['headers']['last-modified'])
+        'timestamp': dateRFC822toUnix(response['headers']['last-modified'])
         })
         try:
             md5 = response['s3cmd-attrs']['md5']
@@ -531,7 +546,7 @@ def compare_filelists(src_list, dst_list, src_remote, dst_remote):
             try:
                 src_md5 = src_list.get_md5(file)
                 dst_md5 = dst_list.get_md5(file)
-            except (IOError,OSError):
+            except (IOError, OSError):
                 # md5 sum verification failed - ignore that file altogether
                 debug(u"IGNR: %s (disappeared)" % (file))
                 warning(u"%s: file disappeared, ignoring." % (file))
@@ -559,12 +574,12 @@ def compare_filelists(src_list, dst_list, src_remote, dst_remote):
     debug("Comparing filelists (direction: %s -> %s)" % (__direction_str(src_remote), __direction_str(dst_remote)))
 
     for relative_file in src_list.keys():
-        debug(u"CHECK: %s" % (relative_file))
+        debug(u"CHECK: '%s'" % relative_file)
 
         if relative_file in dst_list:
             ## Was --skip-existing requested?
             if cfg.skip_existing:
-                debug(u"IGNR: %s (used --skip-existing)" % (relative_file))
+                debug(u"IGNR: '%s' (used --skip-existing)" % relative_file)
                 del(src_list[relative_file])
                 del(dst_list[relative_file])
                 continue
@@ -572,14 +587,14 @@ def compare_filelists(src_list, dst_list, src_remote, dst_remote):
             try:
                 same_file = _compare(src_list, dst_list, src_remote, dst_remote, relative_file)
             except (IOError,OSError):
-                debug(u"IGNR: %s (disappeared)" % (relative_file))
-                warning(u"%s: file disappeared, ignoring." % (relative_file))
+                debug(u"IGNR: '%s' (disappeared)" % relative_file)
+                warning(u"%s: file disappeared, ignoring." % relative_file)
                 del(src_list[relative_file])
                 del(dst_list[relative_file])
                 continue
 
             if same_file:
-                debug(u"IGNR: %s (transfer not needed)" % relative_file)
+                debug(u"IGNR: '%s' (transfer not needed)" % relative_file)
                 del(src_list[relative_file])
                 del(dst_list[relative_file])
 
@@ -592,8 +607,8 @@ def compare_filelists(src_list, dst_list, src_remote, dst_remote):
                 if md5 is not None and md5 in dst_list.by_md5:
                     # Found one, we want to copy
                     dst1 = dst_list.find_md5_one(md5)
-                    debug(u"DST COPY src: %s -> %s" % (dst1, relative_file))
-                    copy_pairs.append((src_list[relative_file], dst1, relative_file))
+                    debug(u"DST COPY src: '%s' -> '%s'" % (dst1, relative_file))
+                    copy_pairs.append((src_list[relative_file], dst1, relative_file, md5))
                     del(src_list[relative_file])
                     del(dst_list[relative_file])
                 else:
@@ -610,12 +625,13 @@ def compare_filelists(src_list, dst_list, src_remote, dst_remote):
             try:
                 md5 = src_list.get_md5(relative_file)
             except IOError:
-               md5 = None
+                md5 = None
             dst1 = dst_list.find_md5_one(md5)
             if dst1 is not None:
                 # Found one, we want to copy
-                debug(u"DST COPY dst: %s -> %s" % (dst1, relative_file))
-                copy_pairs.append((src_list[relative_file], dst1, relative_file))
+                debug(u"DST COPY dst: '%s' -> '%s'" % (dst1, relative_file))
+                copy_pairs.append((src_list[relative_file], dst1,
+                                   relative_file, md5))
                 del(src_list[relative_file])
             else:
                 # we don't have this file, and we don't have a copy of this file elsewhere.  Get it.
